from typing import Any, Awaitable, Dict, Optional

from databases import Database
from graphql import graphql as graphql_exec
from nebulo.gql.alias import Schema
from nebulo.server.jwt import get_jwt_claims_handler
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Route

__all__ = ["get_graphql_route"]


def get_graphql_route(
    gql_schema: Schema,
    database: Database,
    path: str = "/",
    jwt_secret: Optional[str] = None,
    default_role: Optional[str] = None,
    name: Optional[str] = None,
) -> Route:
    """Create a Starlette Route to serve GraphQL requests

    **Parameters**

    * **schema**: _Schema_ = A GraphQL-core schema
    * **database**: _Database_ = Database object for communicating with PostgreSQL
    * **path**: _str_ = URL path to serve GraphQL from, e.g. '/'
    * **jwt_secret**: _str_ = secret key used to encrypt JWT contents
    * **default_role**: _str_ = Default SQL role to use when serving unauthenticated requests
    * **name**: _str_ = Name of the GraphQL serving Starlette route
    """

    get_jwt_claims = get_jwt_claims_handler(jwt_secret)

    async def graphql_endpoint(request: Request) -> Awaitable[JSONResponse]:

        query = await get_query(request)
        variables = await get_variables(request)
        jwt_claims = await get_jwt_claims(request)
        request_context = {
            "request": request,
            "database": database,
            "query": query,
            "variables": variables,
            "jwt_claims": jwt_claims,
            "default_role": default_role,
        }
        result = await graphql_exec(
            schema=gql_schema,
            source=query,
            context_value=request_context,
            variable_values=variables,
        )
        errors = result.errors
        result_dict = {
            "data": result.data,
            "errors": [error.formatted for error in errors or []],
        }
        return JSONResponse(result_dict)

    graphql_route = Route(path=path, endpoint=graphql_endpoint, methods=["POST"], name=name)

    return graphql_route


async def get_query(request: Request) -> Awaitable[str]:
    """Retrieve the GraphQL query from the Starlette Request"""

    content_type = request.headers.get("content-type", "")
    if content_type == "application/graphql":
        return await request.body()
    if content_type == "application/json":
        return (await request.json())["query"]
    raise HTTPException(400, "content-type header must be set")


async def get_variables(request) -> Awaitable[Dict[str, Any]]:
    """Retrieve the GraphQL variables from the Starlette Request"""

    content_type = request.headers.get("content-type", "")
    if content_type == "application/graphql":
        raise NotImplementedError("Getting variables from graphql content type not known")

    if content_type == "application/json":
        return (await request.json()).get("variables", {})
    raise HTTPException(400, "content-type header must be set")



from typing import Optional

from databases import Database
from nebulo.gql.sqla_to_gql import sqla_models_to_graphql_schema
from nebulo.server.exception import http_exception
from nebulo.server.routes import get_graphiql_route, get_graphql_route
from nebulo.sql.reflection.manager import reflect_sqla_models
from sqlalchemy import create_engine
from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware


def create_app(
    connection: str,
    schema: str = "public",
    jwt_identifier: Optional[str] = None,
    jwt_secret: Optional[str] = None,
    default_role: Optional[str] = None,
) -> Starlette:
    """Instantiate the Starlette app"""

    if not (jwt_identifier is not None) == (jwt_secret is not None):
        raise Exception("jwt_token_identifier and jwt_secret must be provided together")

    database = Database(connection)
    # Reflect database to sqla models
    sqla_engine = create_engine(connection)
    sqla_models, sql_functions = reflect_sqla_models(engine=sqla_engine, schema=schema)

    # Convert sqla models to graphql schema
    gql_schema = sqla_models_to_graphql_schema(
        sqla_models,
        sql_functions,
        jwt_identifier=jwt_identifier,
        jwt_secret=jwt_secret,
    )

    graphql_path = "/"

    graphql_route = get_graphql_route(
        gql_schema=gql_schema,
        database=database,
        jwt_secret=jwt_secret,
        default_role=default_role,
        path=graphql_path,
        name="graphql",
    )

    graphiql_route = get_graphiql_route(graphiql_path="/graphiql", graphql_path=graphql_path, name="graphiql")

    _app = Starlette(
        routes=[graphql_route, graphiql_route],
        middleware=[Middleware(CORSMiddleware, allow_origins=["*"])],
        exception_handlers={HTTPException: http_exception},
        on_startup=[database.connect],
        on_shutdown=[database.disconnect],
    )

    return _app



import typing

from sqlalchemy import Text, func, literal, select
from sqlalchemy.sql.selectable import Select


def build_claims(jwt_claims: typing.Dict[str, typing.Any], default_role: typing.Optional[str]) -> Select:
    """Emit statement to set 'jwt.claims.<key>' for each claim in claims dict
    and a 'role'
    """
    # Setting local variables an not be done in prepared statement
    # since JWT claims are signed, literal binds should be ok
    role_key = "role"

    # Set all jwt.claims.*
    claims = [
        func.set_config(
            literal("jwt.claims.").op("||")(func.cast(claim_key, Text())),
            func.cast(str(claim_value), Text()),
            True,
        )
        for claim_key, claim_value in jwt_claims.items()
    ]
    # Set all role claim if exists from jwt
    if role_key in jwt_claims:
        claims.append(
            func.set_config(
                func.cast(role_key, Text()),
                func.cast(str(jwt_claims[role_key]), Text()),
                True,
            )
        )
    # Set default role from config if provided
    elif default_role is not None:
        claims.append(
            func.set_config(
                func.cast(role_key, Text()),
                func.cast(str(default_role), Text()),
                True,
            )
        )
    return select(claims)



import os

from setuptools import find_packages, setup


def read_package_variable(key, filename="__init__.py"):
    """Read the value of a variable from the package without importing."""
    module_path = os.path.join("src/nebulo", filename)
    with open(module_path) as module:
        for line in module:
            parts = line.strip().split(" ", 2)
            if parts[:-1] == [key, "="]:
                return parts[-1].strip("'").strip('"')
    return None


setup(
    name="nebulo",
    version=read_package_variable("VERSION"),
    description="Nebulo: GraphQL API for PostgreSQL",
    author="Oliver Rice",
    author_email="oliver@oliverrice.com",
    license="MIT",
    url="https://github.com/olirice/nebulo",
    project_urls={
        "Documentation": "https://olirice.github.io/nebulo/",
        "Source Code": "https://github.com/olirice/nebulo",
    },
    keywords="graphql sqlalchemy sql api python",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.7",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    entry_points={
        "console_scripts": ["nebulo=nebulo.cli:main", "neb=nebulo.cli:main"],
        "pygments.lexers": ["graphqllexer=nebulo.lexer:GraphQLLexer"],
    },
    install_requires=[
        "aiofiles==0.5.*",
        "appdirs==1.4.3",
        "cachetools==4.0.*",
        "click==7.*",
        "databases[postgresql]==0.2.*",
        "flupy==1.*",
        "graphql-core==3.1.*",
        "inflect==4.1.*",
        "parse==1.15.*",
        "psycopg2-binary==2.8.*",
        "pyjwt==1.7.*",
        "starlette==0.13.*",
        "sqlalchemy==1.3.*",
        "typing-extensions",
        "uvicorn==0.12.*",
    ],
    extras_require={
        "test": ["pytest", "pytest-cov", "requests"],
        "dev": ["pylint", "black", "sqlalchemy-stubs", "pre-commit"],
        "nvim": ["neovim", "python-language-server"],
        "docs": ["mkdocs", "pygments", "pymdown-extensions", "mkautodoc"],
    },
)
